def alpha_diversity_scatter_plot(TaXon_table_xlsx, meta_data_to_test, alpha_w, alpha_h, alpha_s, alpha_font, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import plot, ion, show
    from pathlib import Path
    from matplotlib.pyplot import cm

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    Output_basic_alpha_plot_pdf = Path(str(path_to_outdirs) + "/" + "Alpha_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_scatter_plot.pdf")

    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.PopupError("The meta data has to differ between samples!", title="Error")
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        OTU_abundances_dict = {}
        samples_metadata_list = []

        # remove samples that do not fit the format
        for i, sample in enumerate(samples):
            meta_data = str(Meta_data_table_df.loc[i][meta_data_to_test])
            samples_metadata_list.append([meta_data])

        #################################
        # Calculate Alpha diversity measurements (= observed_otus)

        observed_otus_dict = {}

        for sample in samples:
            observed_otus = len([taxon for taxon in TaXon_table_df[sample].values.tolist() if taxon != 0])
            observed_otus_dict[sample] = observed_otus

        # create list which contains the site for each sample
        categories, color_groups_dict = [], {}

        categories = Meta_data_table_df[meta_data_to_test].values.tolist()

        categories_set = set(categories)

        colors = [plt.cm.tab10(i/float(len(categories_set)-1)) for i in range(len(categories_set))]

        for i, category in enumerate(categories_set):
            color_groups_dict[category] = colors[i]

        ########################################
        # create the plot

        fig, ax = plt.subplots(figsize=(int(alpha_w), int(alpha_h)))
        for i, sample in enumerate(samples):
            group = samples_metadata_list[i][0]
            try:
                group_color = color_groups_dict[group]
            except:
                group_color = color_groups_dict[int(group)]
            plt.scatter(sample, observed_otus_dict[sample], s=int(alpha_s), label=group, c=[group_color])
        plt.xticks(fontsize=alpha_font, rotation=90)
        plt.yticks(fontsize=alpha_font)
        plt.title("Observed OTUs", fontsize=alpha_font)
        plt.ylabel("# OTUs", fontsize=alpha_font, rotation=90)
        ax.set_axisbelow(True)
        ax.grid(color='gray', alpha=0.1)

        # add a legend
        handles, labels = ax.get_legend_handles_labels()
        unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
        plt.legend(*zip(*unique), loc='center left', bbox_to_anchor=(1, 0.5))

        plt.draw()
        plt.pause(0.001)
        answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
        if answer == "Yes":
            plt.savefig(Output_basic_alpha_plot_pdf, bbox_inches='tight')
            plt.close()
            print("\n" + "Alpha diversity estimate plots are found in", path_to_outdirs, "/Alpha_diversity/")
            sg.Popup("Alpha diversity estimate are found in", path_to_outdirs, "/Alpha_diversity/", title="Finished", keep_on_top=True)

            from taxontabletools.create_log import ttt_log
            ttt_log("alpha diversity scatter", "analysis", TaXon_table_xlsx.name, Output_basic_alpha_plot_pdf.name, meta_data_to_test, path_to_outdirs)

        else:
            plt.close()

    else:
        sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)
        print("Error: The samples between the taxon table and meta table do not match!")

def alpha_diversity_boxplot(TaXon_table_xlsx, meta_data_to_test, alpha_w, alpha_h, alpha_s, alpha_font, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import plot, ion, show
    from pathlib import Path
    from matplotlib.pyplot import cm

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    Output_basic_alpha_plot_pdf = Path(str(path_to_outdirs) + "/" + "Alpha_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_boxplot.pdf")

    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.PopupError("The meta data has to differ between samples!", title="Error")
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        OTU_abundances_dict = {}
        samples_metadata_list = []

        # remove samples that do not fit the format
        for i, sample in enumerate(samples):
            meta_data = str(Meta_data_table_df.loc[i][meta_data_to_test])
            samples_metadata_list.append(meta_data)

        #################################
        # Calculate Alpha diversity measurements (= observed_otus)

        observed_otus_dict = {}

        for i, sample in enumerate(samples):
            observed_otus = len([taxon for taxon in TaXon_table_df[sample].values.tolist() if taxon != 0])
            categorie = samples_metadata_list[i]
            if categorie not in observed_otus_dict.keys():
                observed_otus_dict[categorie] = [observed_otus]
            else:
                observed_otus_dict[categorie] = observed_otus_dict[categorie] + [observed_otus]

        # create boxplot
        labels, data = np.array(list(observed_otus_dict.keys()), dtype=object), np.array(list(observed_otus_dict.values()), dtype=object)
        plt.figure(figsize=(int(alpha_w), int(alpha_h)))
        plt.grid(color='gray', alpha=0.1)
        plt.boxplot(data,
            patch_artist=True,
            boxprops=dict(facecolor="#0000000a", color="black"),
            capprops=dict(color="black"),
            whiskerprops=dict(color="black"),
            flierprops=dict(color="black", markeredgecolor="black"),
            medianprops=dict(color="blue"),
            )
        plt.xticks(range(1, len(labels) + 1), labels, fontsize=alpha_font)
        plt.yticks(fontsize=alpha_font)
        plt.title("Observed OTUs", fontsize=alpha_font)
        plt.ylabel("# OTUs", fontsize=alpha_font, rotation=90)

        plt.draw()
        plt.pause(0.001)
        answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
        if answer == "Yes":
            plt.savefig(Output_basic_alpha_plot_pdf, bbox_inches='tight')
            plt.close()
            print("\n" + "Alpha diversity estimate plots are found in", path_to_outdirs, "/Alpha_diversity/")
            sg.Popup("Alpha diversity estimate are found in", path_to_outdirs, "/Alpha_diversity/", title="Finished", keep_on_top=True)

            from taxontabletools.create_log import ttt_log
            ttt_log("alpha diversity boxplot", "analysis", TaXon_table_xlsx.name, Output_basic_alpha_plot_pdf.name, meta_data_to_test, path_to_outdirs)

        else:
            plt.close()

    else:
        sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)
        print("Error: The samples between the taxon table and meta table do not match!")



def read_proportions(TaXon_table_xlsx, taxonomic_level, path_to_outdirs, x_read_props_scatter, y_read_props_scatter, font_read_props_scatter, no_hit_visualization, read_props_alternating_colors):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import plot, ion, show
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    TaXon_table_df = TaXon_table_df.replace(np.nan, 'nan', regex=True)
    samples_list = TaXon_table_df.columns.tolist()[10:]
    Species_read_proportion_dict = {}

    # check for presence absence data
    # otherwise abort and print error message
    pa_test = set([val for sublist in TaXon_table_df[samples_list].values.tolist() for val in sublist])
    if pa_test == {1,0}:
        print("Please do not use presence absence data!")
        sg.Popup("Please do not use presence absence data!", title=("Error"))
        raise RuntimeError

    ##############################################################################
    # scatter plot

    read_proportion_plot_directory = Path(str(path_to_outdirs) + "/" + "Read_proportions_plots" + "/" + TaXon_table_xlsx.stem)
    scatter_pdf = Path(str(read_proportion_plot_directory) + "_" + taxonomic_level + "_scatter.pdf")

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(samples_list) + 1
    ############################################################################

    for sample in samples_list:

        # check if the option "best hit" is chosen
        if no_hit_visualization == "best_hit":
            TaXon_table_list = []

            taxon_levels_dict = {"Phylum": 1, "Class": 2, "Order": 3, "Family": 4, "Genus": 5, "Species": 6}
            no_taxonomic_level = taxon_levels_dict[taxonomic_level]
            df = TaXon_table_df[['IDs', "Phylum", "Class", "Order", "Family", "Genus", "Species", sample]]
            df_2 = df[[sample]]/df[[sample]].sum()
            df = df.assign(perc=df_2.values)

            # loop through the OTU dataframe and create new dict to store the taxonomy and read counts for each OTU
            for OTU in df.values.tolist():
                OTU_dict = {}
                # if the taxonomic level that was chosen by the user is "nan" look for the best hit
                if OTU[no_taxonomic_level] == "nan":
                    test_taxonomy_list = []
                    # loop through all taxonomic levels and check were the first level was identified (== is not "nan")
                    # append the best hit to a list
                    for test in range(1,no_taxonomic_level-1):
                        if OTU[test] != "nan":
                            test_taxonomy_list.append([test, OTU[test]])
                    # collect the OTU ID, the taxon name and the read proportions
                    TaXon_table_list.append([OTU[0], OTU[test_taxonomy_list[-1][0]] + "*", OTU[-1]])
                # skip it the OTU was identified on the correct level
                else:
                    TaXon_table_list.append([OTU[0], OTU[no_taxonomic_level], OTU[-1]])

        else:
            df = TaXon_table_df[['IDs', taxonomic_level, sample]]
            df_2 = df[[sample]]/df[[sample]].sum()
            df = df.assign(perc=df_2.values)
            df = df.drop(columns=[sample])
            TaXon_table_list = df.values.tolist()

        for line in TaXon_table_list:

            species = line[1]

            reads = line[2] * 100
            dict_key = sample, species

            if dict_key not in Species_read_proportion_dict.keys():
                Species_read_proportion_dict[dict_key] = reads
            else:
                Species_read_proportion_dict[dict_key] = Species_read_proportion_dict[dict_key] + reads

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            print('Cancel')
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    sample_list = []
    species_list = []
    read_proportion_list = []

    for dict_key, value in list(Species_read_proportion_dict.items()):
        sample_list.append(dict_key[0])
        species_list.append(dict_key[1])
        read_proportion_list.append(value)

    # create data
    x = sample_list #['a', 'b', 'c']
    y = species_list #['d', 'f', 'g']-
    z = read_proportion_list #[199, 290, 290]

    # calculate the number of taxa
    n_taxa = len(set(species_list))
    # calculate the number of scatter entries for the plot
    n_species_list = len(species_list)

    # create an even color list if the number of taxa is even
    if (n_taxa % 2) == 0:
       clist = (["#52af2fff", "#4472c4ff"] * int(n_species_list / 2))
    # create an odd color list if the number of taxa is odd
    # this every sample will start with the same color
    else:
       clist = (["#52af2fff", "#4472c4ff"] * (int((n_taxa - 1) / 2)) + ["#52af2fff"]) * n_species_list
       clist = clist[:n_species_list]

    title = "Read proportions (" + taxonomic_level + ")"
    fig, ax = plt.subplots(figsize=(int(x_read_props_scatter), int(y_read_props_scatter)))
    if read_props_alternating_colors == True:
        scatter = ax.scatter(x, y, s=z, color=clist, alpha=0.9)
    else:
        scatter = ax.scatter(x, y, s=z, alpha=0.9)
    plt.xticks(fontsize=font_read_props_scatter, rotation=90)
    plt.yticks(fontsize=font_read_props_scatter)
    plt.title(title, fontsize=8)
    ax.set_axisbelow(True)
    ax.grid(color='gray', alpha=0.1)

    # produce a legend with a cross section of sizes from the scatter
    handles, labels = scatter.legend_elements(prop="sizes", alpha=0.1, color="blue")
    legend2 = ax.legend(handles, labels, loc='center left', bbox_to_anchor=(1, 0.5), title="Proportion(%)")

    ########################

    plt.show(block=False)
    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
    if answer == "Yes":
        print("Saving...")
        plt.savefig(scatter_pdf)
        plt.close()
        closing_text = "Read proportion plot is found under:\n" + '/'.join(str(scatter_pdf).split("/")[-4:])
        print(closing_text)
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        from taxontabletools.create_log import ttt_log
        ttt_log("read proportions", "analysis", TaXon_table_xlsx.name, scatter_pdf.name, no_hit_visualization, path_to_outdirs)

    else:
        plt.close()



def pie_charts(TaXon_table_xlsx, path_to_outdirs, pc_label_font_size):

    ## import functions
    import os
    import PySimpleGUI as sg
    from pathlib import Path
    import pandas as pd
    import numpy as np
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 167*2
    ############################################################################

    TaXon_table_file =  Path(TaXon_table_xlsx)
    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_file)
    allowed_taxa = ["A_Phylum","B_Class","C_Order","D_Family","E_Genus", "F_Species"]

    print("\n" + "Input file:", TaXon_table_file.name)

    for taxon_query in allowed_taxa:
        taxon_level = taxon_query
        taxon_query = taxon_query[2:]

        print("Calclulating pie chart for:", taxon_query, "level")

        dirName = Path(str(path_to_outdirs) + "/" + "Pie_charts" + "/" + str(TaXon_table_file.stem))

        if not os.path.exists(dirName):
            os.mkdir(dirName)

        Output_name = Path(str(dirName) + "/" + taxon_level + ".pdf")


        with PdfPages(Output_name) as pdf:

            taxonomy_results_xlsx = pd.ExcelFile(TaXon_table_file)
            data = pd.read_excel(taxonomy_results_xlsx, 'TaXon table', header=0)

            taxa = data[taxon_query].values.tolist()
            taxa_unique = list(dict.fromkeys(taxa))

            taxa_labels = []
            taxa_sizes = []

            for taxon_name in taxa_unique:
                taxa_labels.append(str(taxon_name))
                taxa_sizes.append(taxa.count(taxon_name))

            patches, texts, autotexts = plt.pie(taxa_sizes, startangle=140, autopct='%1.1f%%', wedgeprops=dict(width=0.5))

            bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.12)
            kw = dict(arrowprops=dict(arrowstyle='-', linewidth=0.1), bbox=bbox_props, zorder=0, va="center")

            for i, p in enumerate(patches):
                ang = (p.theta2 - p.theta1)/2. + p.theta1
                y = np.sin(np.deg2rad(ang))
                x = np.cos(np.deg2rad(ang))
                horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
                connectionstyle = "angle,angleA=0,angleB={}".format(ang)
                kw["arrowprops"].update({"connectionstyle": connectionstyle})
                plt.annotate(taxa_labels[i], xy=(x, y), xytext=(1.35*np.sign(x), 1.4*y), horizontalalignment=horizontalalignment, **kw, fontsize=pc_label_font_size)

            for i in range(len(texts)):
                texts[i].set_fontsize(4)

            for i in range(len(autotexts)):
                autotexts[i].set_fontsize(4)

            plt.axis('equal')
            plt.title("OTU abundance on " + taxon_query + " level", fontsize=10, pad=20)
            fileName= TaXon_table_file.name
            plt.suptitle(fileName, x=0.2, y=0.05, fontsize=5)

            pdf.savefig()
            plt.close()

            #########################

            taxa_labels = []
            taxa_sizes = []

            for taxon_name in taxa_unique:
                if "nan" != str(taxon_name):
                    taxa_labels.append(str(taxon_name))
                    taxa_sizes.append(taxa.count(taxon_name))

            patches, texts, autotexts = plt.pie(taxa_sizes, startangle=140, autopct='%1.1f%%', wedgeprops=dict(width=0.5))

            bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.12)
            kw = dict(arrowprops=dict(arrowstyle='-', linewidth=0.1), bbox=bbox_props, zorder=0, va="center")

            for i, p in enumerate(patches):
                ang = (p.theta2 - p.theta1)/2. + p.theta1
                y = np.sin(np.deg2rad(ang))
                x = np.cos(np.deg2rad(ang))
                horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
                connectionstyle = "angle,angleA=0,angleB={}".format(ang)
                kw["arrowprops"].update({"connectionstyle": connectionstyle})
                plt.annotate(taxa_labels[i], xy=(x, y), xytext=(1.35*np.sign(x), 1.4*y), horizontalalignment=horizontalalignment, **kw, fontsize=pc_label_font_size)

            for i in range(len(texts)):
                texts[i].set_fontsize(4)

            for i in range(len(autotexts)):
                autotexts[i].set_fontsize(4)

            plt.axis('equal')
            plt.title("OTU abundance (excluding nan) on " + taxon_query + " level", fontsize=10, pad=20)

            pdf.savefig()

            if taxon_query == "Species":
                plt.show(block=False)
                closing_text = "\n" + "Pie charts are found in: " + str(path_to_outdirs) + "/Pie_charts/"
                print(closing_text)
                sg.Popup(closing_text, title="Finished", keep_on_top=True)

            plt.close()

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            print('Cancel')
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_bar.UpdateBar(progress_update)
        progress_update += 167
        ############################################################################
    window_progress_bar.close()

    from taxontabletools.create_log import ttt_log
    placeholder = TaXon_table_file.name + " (multiple pie charts)"
    ttt_log("pie charts", "analysis", TaXon_table_file.name, placeholder, "nan", path_to_outdirs)



def PCoA_analysis(TaXon_table_xlsx, meta_data_to_test, pcoa_w, pcoa_h, pcoa_s, path_to_outdirs):

    import pandas as pd
    import numpy as np
    from skbio.diversity import beta_diversity
    from skbio.stats.ordination import pcoa
    from skbio.stats.distance import anosim
    import matplotlib.pyplot as plt
    from pathlib import Path
    import PySimpleGUI as sg

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
    output_pdf = Path(str(path_to_outdirs) + "/" + "PCoA_plots" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + ".pdf")
    output_xlsx = Path(str(path_to_outdirs) + "/" + "PCoA_plots" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + ".xlsx")

    # check for presence absence data
    # otherwise abort and print error message
    pa_test = set([val for sublist in TaXon_table_df[TaXon_table_samples].values.tolist() for val in sublist])
    if pa_test != {1,0}:
        print("Please use presence absence data!")
        sg.Popup("Please use presence absence data!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == len(Meta_data_table_df['Samples'].tolist()):
        print("The meta data is unique for all samples. Please adjust the meta data table!")
        sg.Popup("The meta data is unique for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        print("The meta data is similar for all samples. Please adjust the meta data table!")
        sg.Popup("The meta data is similar for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        data = TaXon_table_df[TaXon_table_samples].transpose().values.tolist()
        jc_dm = beta_diversity("jaccard", data, samples)
        ordination_result = pcoa(jc_dm)
        metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()

        anosim_results = anosim(jc_dm, metadata_list, permutations=999)
        anosim_r = round(anosim_results['test statistic'], 5)
        anosim_p = anosim_results['p-value']
        textbox = meta_data_to_test + "\nAnosim " + "R = " + str(anosim_r) + " " + "p = " + str(anosim_p)

        #######################################################################################
        # create window to ask for PCoA axis to test
        def slices(list, slice):
            for i in range(0, len(list), slice):
                yield list[i : i + slice]

        # collect the PCoA proportion explained values
        proportion_explained_list = []
        for i, pcoa_axis in enumerate(ordination_result.proportion_explained):
            proportion_explained_list.append("PC" + str(i+1) + " (" + str(round(pcoa_axis* 100, 2)) + " %)")

        pcoa_axis_checkboxes = list(slices([sg.Checkbox(name, key=name, size=(15,1)) for name in proportion_explained_list], 4))

        pcoa_window_layout = [
                    [sg.Text('Check two axis to be displayed')],
                    [sg.Frame(layout = pcoa_axis_checkboxes, title = '')],
                    [sg.Button('Plot', key='Plot'), sg.Checkbox("Invert axes", key="inverted_axis")],
                    [sg.Button('Back')],
                    ]

        pcoa_window = sg.Window('PCoA axis', pcoa_window_layout, keep_on_top=True)

        while True:
            event, values = pcoa_window.read()

            if event == 'Plot':
                # collect the pcoa axis values
                axis_to_plot = [key for key,value in values.items() if value == True and "PC" in key]
                # pass on only if two pcoa axes were checked
                if len(axis_to_plot) != 2:
                    sg.Popup("Please choose exactly two PCoA axes", title="Error", keep_on_top=True)
                else:
                    if values["inverted_axis"] == True:
                        cat1 = axis_to_plot[0].split()[0]
                        cat2 =  axis_to_plot[1].split()[0]
                    else:
                        cat2 = axis_to_plot[0].split()[0]
                        cat1 =  axis_to_plot[1].split()[0]
                    df_pcoa = ordination_result.samples[[cat1, cat2]]
                    df_pcoa.insert(2, "label", Meta_data_table_df[meta_data_to_test].values.tolist(), True)
                    plt.figure(figsize=(int(pcoa_w), int(pcoa_h)))
                    plt.grid(color='gray', alpha=0.1)
                    groups = df_pcoa.groupby("label")
                    for name, group in groups:
                        plt.plot(group[cat1], group[cat2], marker="o", linestyle="", label=name, ms=pcoa_s)
                    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
                    plt.xlabel(cat1 + " (" + str(round(ordination_result.proportion_explained[cat1]* 100, 2)) + " %)")
                    plt.ylabel(cat2 + " (" + str(round(ordination_result.proportion_explained[cat2]* 100, 2)) + " %)")
                    plt.title(textbox)

                    plt.show(block=False)
                    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
                    if answer == "Yes":
                        plt.savefig(output_pdf)
                        plt.close()
                        ordination_result.samples[[cat1, cat2]].to_excel(output_xlsx)
                        closing_text = "\n" + "PCoA plots are found in: " + str(path_to_outdirs) + "/PCoA_plots/"
                        print(closing_text)
                        sg.Popup(closing_text, title="Finished", keep_on_top=True)

                        from taxontabletools.create_log import ttt_log
                        ttt_log("pcoa analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)

                        break
                    else:
                        plt.close()

            if event == 'Back':
                break

        pcoa_window.close()



## TaXon table filtering
#1.a create mask for user input
def create_taxa_mask(TaXon_table_xlsx, mask):

    import pandas as pd
    from pandas import DataFrame
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    data = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)

    available_taxa = set(data[mask].values.tolist())
    available_taxa = [x for x in available_taxa if str(x) != 'nan']
    available_taxa = sorted(list(available_taxa))

    return available_taxa

# 1.b filter taxa from taxon table
def taxon_table_filter(TaXon_table_xlsx, exclude_taxa, mask, appendix_name, threshold, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    from pathlib import Path

    TaXon_table_file =  Path(TaXon_table_xlsx)

    print("\n" + "Input file:", TaXon_table_file.name)

    # convert taxa to exclude to a list if only one taxon is given (which is then string)
    if type(exclude_taxa) == str:
        exclude_taxa = [exclude_taxa]

    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_file)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)

    # check for taxa to filter
    mask_position = list(df.columns).index(mask)
    df_columns = df.columns

    rows_to_keep = []

    df_rows = df.values.tolist()
    for row in df_rows:
        taxon_to_evaluate = row[mask_position]
        if taxon_to_evaluate not in exclude_taxa:
            if str(taxon_to_evaluate) != 'nan':
                rows_to_keep.append(row)

    df_out = pd.DataFrame(rows_to_keep)

    similarity_position = list(df_columns).index("Similarity")
    threshold = int(threshold)

    filtered_rows = []

    for index, row in df_out.iterrows():
        similarity = list(row)[similarity_position]
        if similarity != 'No Match':
            if int(similarity) >= threshold:
                filtered_rows.append(list(row))

    df_out = pd.DataFrame(filtered_rows)

    if df_out.empty:
        sg.PopupError('Filter theshold were to harsh: Nothing to print', title="Error", keep_on_top=True)
        print('\n' + 'Filter theshold were to harsh: Nothing to print', '\n')

    else:
        df_out.columns = df_columns

        # write output file
        file_name = TaXon_table_file.stem
        output_name = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + file_name + "_" + appendix_name + ".xlsx")
        threshold_output = "Similarity threshold = " + str(threshold)
        exclude_taxa.append(threshold_output)
        df_exclude_taxa = pd.DataFrame(exclude_taxa)
        df_exclude_taxa.columns = ['Filter criteria']
        writer = pd.ExcelWriter(output_name, engine = 'xlsxwriter')
        df_out.to_excel(writer, sheet_name = 'TaXon table', index=False)
        df_exclude_taxa.to_excel(writer, sheet_name = 'Filter criteria', index=False)
        writer.save()
        writer.close()

        closing_text = "Taxon table is found under:\n" + '/'.join(str(output_name).split("/")[-4:])
        print(closing_text)
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        from taxontabletools.create_log import ttt_log
        ttt_log("taxon filter", "processing", TaXon_table_file.name, output_name.name, "nan", path_to_outdirs)

# 2.a create mask for user input
def create_sample_mask(TaXon_table_xlsx):

    import pandas as pd
    from pandas import DataFrame
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)

    available_samples = df.columns.tolist()[10:]

    return available_samples

# 2.b filter samples from taxon list
def exclude_samples(TaXon_table_xlsx, exluded_samples, appendix_name, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    from pathlib import Path

    TaXon_table_file =  Path(TaXon_table_xlsx)

    print("\n" + "Input file:", TaXon_table_file.name)

    TaXon_table_xlsx_path = TaXon_table_xlsx
    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)

    if type(exluded_samples) == str:
        exluded_samples = [exluded_samples]

    for sample in exluded_samples:
        df = df.drop(sample, axis=1)

    header = df.columns.values.tolist()

    row_filter_list = []

    for row in df.values.tolist():
        reads = set(row[10:])
        if reads != {0}:
            row_filter_list.append(row)

    df = pd.DataFrame(row_filter_list)
    df.columns = header

    file_name = TaXon_table_file.stem
    output_name = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + file_name + "_" + appendix_name + ".xlsx")
    df.to_excel(output_name, sheet_name = 'TaXon table', index=False)

    closing_text = "Taxon table is found under:\n" + '/'.join(str(output_name).split("/")[-4:])
    print(closing_text)
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("sample filter", "processing", TaXon_table_file.name, output_name.name, "nan", path_to_outdirs)



def site_occupancy(TaXon_table_xlsx, meta_data_to_test, taxonomic_level, path_to_outdirs, x_site_occ, y_site_occ, font_site_occ):

    import os
    import pandas as pd
    from pandas import DataFrame
    from pathlib import Path
    import matplotlib.pyplot as plt
    import PySimpleGUI as sg

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header = 0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    TaXon_table_n_samples = len(TaXon_table_samples)

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    answer = "Ask"
    output_message = "No"

    try:
        if os.path.exists(Meta_data_table_xlsx):

            Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header = 0)
            Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

            n_sites = len(set(Meta_data_table_df[meta_data_to_test].tolist()))

            if (sorted(TaXon_table_samples) == sorted(Meta_data_table_samples) and TaXon_table_n_samples != n_sites):

                site_occupancy_dict = {}

                sites = set(Meta_data_table_df[meta_data_to_test].tolist())

                for site in sites:
                    # this can either be a species name or the above specified taxonomic level
                    present_OTU_list = []

                    # extract samples that belong to the site from the metadata file
                    included_samples_list = Meta_data_table_df[Meta_data_table_df.values  == site]['Samples'].values.tolist()

                    # count the number of samples per site to calculate the site occupancy
                    n_samples = len(included_samples_list)

                    # create a list of all species (or the specified taxonomic level)
                    if taxonomic_level == "OTUs":
                        taxonomic_level = "IDs"
                    overall_included_species_list = TaXon_table_df[taxonomic_level].values.tolist()
                    # make the list unique
                    overall_included_species_set = set(overall_included_species_list)
                    # remove potential 'nan's from the list
                    overall_included_species_set = [x for x in overall_included_species_set if str(x) != 'nan']

                    # create a set of species that is present at the sites
                    for sample in included_samples_list:

                        OTUs_per_species_list = []

                        # check the read abundaces for each sample
                        read_abundace_list = TaXon_table_df[sample].values.tolist()

                        # enumerate the read abundaces for each sample and collect all lines that have more than one read
                        for i, read_abundance in enumerate(read_abundace_list):
                            species = TaXon_table_df[taxonomic_level][i]
                            # if reads are present, collect the species name (or the specified taxonomic level) from the TaXon table
                            if read_abundance != 0:
                                OTUs_per_species_list.append(species)

                        # remove all nans
                        OTUs_per_species_list = [x for x in OTUs_per_species_list if str(x) != 'nan']
                        # make list unique
                        OTUs_per_species_list = list(set(OTUs_per_species_list))
                        # append to list of species for the current site
                        present_OTU_list.append(OTUs_per_species_list)

                    # flatten the list of present species per site
                    present_OTU_list_flattened = [val for sublist in present_OTU_list for val in sublist]

                    # store occupancy of each species in a dict, will be accessed by position in list
                    occupancy_dict = {}

                    # count the number of occurences for each species and calculate the occpancy based on the number of samples
                    for species in overall_included_species_set:
                        count = present_OTU_list_flattened.count(species)
                        occupancy = count / n_samples
                        occupancy_dict[species] = occupancy

                    occupancy_dict = {k: v for k, v in sorted(occupancy_dict.items(), key=lambda item: item[1])}
                    occupancy_list = list(occupancy_dict.values())
                    species_list = list(occupancy_dict.keys())

                    # create a simple bar plot for each site
                    plt.figure(figsize=(int(x_site_occ), int(y_site_occ)))
                    plt.bar(species_list, occupancy_list)
                    plt.title(site + " (" + taxonomic_level + ")")
                    plt.xticks(fontsize=font_site_occ, rotation=90)
                    plt.yticks(fontsize=font_site_occ)
                    plt.ylabel("occupancy", fontsize=font_site_occ)
                    plt.ylim(0,1)

                    occupancy_plot_directory = Path(str(path_to_outdirs) + "/" + "Site_occupancy_plots" + "/" + TaXon_table_xlsx.stem)
                    if not os.path.exists(occupancy_plot_directory):
                        os.mkdir(occupancy_plot_directory)

                    occupancy_plot_pdf = Path(str(occupancy_plot_directory) + "/" + site + "_" + taxonomic_level + ".pdf")
                    occupancy_table = Path(str(occupancy_plot_directory) + "/" + site + "_" + taxonomic_level + ".xlsx")

                    if answer == "Ask":
                        # show the first plot
                        plt.show(block=False)
                        # ask for answer
                        answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
                        # if answer is yes, save all upcoming plots
                        if answer == "Yes":
                            output_message = "Yes"
                            occupancy_df = pd.DataFrame(occupancy_list, species_list)
                            occupancy_df.columns = ["Occupancy"]
                            occupancy_df.index.name = "Taxon"
                            occupancy_df = occupancy_df.sort_values("Occupancy")
                            # sort the table numerical if OTUs were chosen
                            if taxonomic_level == "IDs":
                                sort_list = []
                                for OTU in occupancy_df.index.tolist():
                                    sort_list.append(int(OTU.split("_")[1]))
                                occupancy_df["sort"] = sort_list
                                occupancy_df = occupancy_df.sort_values("sort")
                                occupancy_df = occupancy_df.drop("sort", axis=1)
                            occupancy_df.to_excel(occupancy_table)
                            # save the plot
                            plt.savefig(occupancy_plot_pdf)
                            plt.close()
                        else:
                            # answer is "No"
                            plt.close()
                    elif answer == "Yes":
                        # write the other sites
                            occupancy_df = pd.DataFrame(occupancy_list, species_list)
                            occupancy_df.columns = ["Occupancy"]
                            occupancy_df.index.name = "Taxon"
                            occupancy_df = occupancy_df.sort_values("Occupancy")
                            # sort the table numerical if OTUs were chosen
                            if taxonomic_level == "IDs":
                                sort_list = []
                                for OTU in occupancy_df.index.tolist():
                                    sort_list.append(int(OTU.split("_")[1]))
                                occupancy_df["sort"] = sort_list
                                occupancy_df = occupancy_df.sort_values("sort")
                                occupancy_df = occupancy_df.drop("sort", axis=1)
                            occupancy_df.to_excel(occupancy_table)
                            # save the plot
                            plt.savefig(occupancy_plot_pdf)
                            plt.close()
                    else:
                            plt.close()

                if output_message == "Yes":
                    closing_text = "Site occupancy plots are found under:\n" + '/'.join(str(occupancy_plot_pdf).split("/")[-4:])
                    print(closing_text)
                    sg.Popup(closing_text, title="Finished", keep_on_top=True)

                    from taxontabletools.create_log import ttt_log
                    placeholder = TaXon_table_xlsx.name + " (multiple site occupancy plots)"
                    ttt_log("site occupancy", "analysis", TaXon_table_xlsx.name, placeholder, meta_data_to_test, path_to_outdirs)

            else:
                sg.PopupError("Please check your Metadata file and Taxon table file: The samples do not match or the metadata is unique for all samples!", keep_on_top=True)
                print("Please check your metadata file and Taxon table file: The samples do not match or the metadata is unique for all samples!")

        else:
            sg.PopupError("Missing metadata file!", keep_on_top=True)
            print("Missing metadata file!")

    except:
        exception_text = "Something went wrong!" + "\n" + "Do not use numbers as metadata keys." + "\n" + "Please check your metadata file and read the manual."
        sg.PopupError(exception_text, title="Error", keep_on_top=True)
        print(exception_text)



def rarefaction_curve_OTUs(TaXon_table_xlsx, replicates, error_style, rarefaction_ylim, path_to_outdirs):

    import random
    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import plot, ion, show
    from pathlib import Path

    TaXon_table_file = Path(TaXon_table_xlsx)

    print("\n" + "Input file:", TaXon_table_file.name)

    output_file = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_otus.pdf")

    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)

    available_samples = df.columns.tolist()[10:]
    sample_dict_clean = {}

    # iterate through all available samples
    for sample in available_samples:
        # create a dict for the read numbers of the respective sample
        sample_dict = df[sample].to_dict()
        # create an empty list for keys to store
        key_list = []
        # iterate through the read table of the sample
        for key, value in sample_dict.items():
            # if the read number of the OTU is NOT ZERO
            if value != 0:
                # append the OTU number to the key list
                key_list.append(key)
        # add the key list of each sample to a clean dictionary
        sample_dict_clean[sample] = key_list

    # draw once for each sample
    number_of_draws = len(sample_dict_clean.keys())

    # dictionary to store the drawing results
    draw_dictionary = {}

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / replicates
    ############################################################################

    for n_reps in range(0, replicates):
        # store the original dictionary to start over again
        # a copy of the original dictionary is required, because the samples will be removed with each draw
        # thus for each replicate a new dictionary to draw from has to be created
        sample_dict_to_draw = dict(sample_dict_clean)

        OTU_list = []
        OTU_set = []

        for i in range(0, number_of_draws):
            # choose a random sample from the dictionary
            random_choice = random.choice(list(sample_dict_to_draw.keys()))
            # extract the OTU IDs from the chosen sample and add them to the already existing OTU IDs
            OTU_list = OTU_list + sample_dict_clean[random_choice]
            # create a unique set
            OTU_set = set(OTU_list)
            # number of OTUs
            n_OTUs = len(OTU_set)
            # now add the unique OTU list to the output dictionary
            # if the key is not in the dict, create a new entry (= OTU ID plus number of OTUs)
            if i not in draw_dictionary.keys():
                draw_dictionary[i] = [n_OTUs]
            # if the key already exists, calculate the sum of the already existing number of OTUs and the new number of OTUs
            else:
                # create a new list to store the current number of OTUs
                add_OTUs_list = draw_dictionary[i]
                add_OTUs_list.append(n_OTUs)
                draw_dictionary[i] = add_OTUs_list

            # remove the sample to draw only once
            sample_dict_to_draw.pop(random_choice)

        ############################################################################
        event, values = window_progress_bar.read(timeout=1)
        if event == 'Cancel'  or event is None:
            print('\nCancel')
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    # create a dict to store the average number of OTUs per draw
    rarefaction_dict_average, rarefaction_dict_stdef = {}, {}

    def average(lst):
        return sum(lst) / len(lst)

    # iterate through the draw_dictionary and calculate the average number of OTUs
    for key, value in draw_dictionary.items():
        average_OTUs = average(draw_dictionary[key])
        stdef_OTUs = np.std(draw_dictionary[key], dtype=np.float64)
        rarefaction_dict_average[key] = average_OTUs
        rarefaction_dict_stdef[key] = stdef_OTUs

    if error_style == "a":
        draws = [i+1 for i in rarefaction_dict_average.keys()] #list(rarefaction_dict_average.keys())
        n_OTUs = list(rarefaction_dict_average.values())
        error_bar = list(rarefaction_dict_stdef.values())
        plt.figure(figsize=(20, 10))
        if rarefaction_ylim != '':
            plt.ylim(0, int(rarefaction_ylim))
        plt.errorbar(draws, n_OTUs, error_bar, linewidth=0.8, color='blue', capsize=1.3, capthick=1, ecolor='lightgrey')
        plt.xticks(np.arange(1, len(draws)+1, step=1))
        plt.xlabel('# samples')
        plt.ylabel('# OTUs')
        plt.title('repetitions = ' + str(replicates))


    elif error_style == "b":
        draws = [i+1 for i in rarefaction_dict_average.keys()] #list(rarefaction_dict_average.keys())
        n_OTUs = list(rarefaction_dict_average.values())
        y = np.asarray(n_OTUs)
        error_bar = np.asarray(list(rarefaction_dict_stdef.values()))
        plt.figure(figsize=(20, 10))
        if rarefaction_ylim != '':
            plt.ylim(0, int(rarefaction_ylim))
        plt.plot(draws, n_OTUs)
        plt.fill_between(draws, y-error_bar, y+error_bar, alpha=0.1)
        plt.xticks(np.arange(1, len(draws)+1, step=1))
        plt.xlabel('# samples')
        plt.ylabel('# OTUs')
        plt.title('repetitions = ' + str(replicates))


    plt.show(block=False)
    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
    if answer == "Yes":
        plt.savefig(output_file, bbox_inches='tight')
        plt.close()
        closing_text = "\n" + "Rarefaction curves are found in: " + str(path_to_outdirs) + "/rarefaction_curves/"
        print(closing_text)
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        from taxontabletools.create_log import ttt_log
        ttt_log("rarefaction curve OTUs", "analysis", TaXon_table_file.name, output_file.name, "nan", path_to_outdirs)

    else:
        plt.close()


def rarefaction_curve_species(TaXon_table_xlsx, replicates, error_style, rarefaction_ylim, path_to_outdirs):

    import random
    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import plot, ion, show
    from pathlib import Path

    TaXon_table_file = Path(TaXon_table_xlsx)

    print("\n" + "Input file:", TaXon_table_file.name)

    output_file = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_species.pdf")

    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)
    df = df.replace(np.nan,"nan")

    available_samples = df.columns.tolist()[10:]
    sample_dict_clean = {}

    # iterate through all available samples
    for sample in available_samples:
        # create a dict for the read numbers of the respective sample for each species
        sample_OTU_list = df[[sample, "Species"]].values.tolist()
        # select only the present Species
        sample_species_list = list(set([OTU[1] for OTU in sample_OTU_list if (OTU[0] != 0 and OTU[1] != "nan")]))
        # store the species in a dictionary
        sample_dict_clean[sample] = sample_species_list

    # draw once for each sample
    number_of_draws = len(sample_dict_clean.keys())

    # dictionary to store the drawing results
    draw_dictionary = {}

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / replicates
    ############################################################################

    for n_reps in range(0, replicates):
        # store the original dictionary to start over again
        # a copy of the original dictionary is required, because the samples will be removed with each draw
        # thus for each replicate a new dictionary to draw from has to be created
        sample_dict_to_draw = dict(sample_dict_clean)

        species_list = []
        species_set = []

        for i in range(0, number_of_draws):
            # choose a random sample from the dictionary
            random_choice = random.choice(list(sample_dict_to_draw.keys()))
            # extract the OTU IDs from the chosen sample and add them to the already existing OTU IDs
            species_list = species_list + sample_dict_clean[random_choice]
            # create a unique set
            species_set = set(species_list)
            # number of OTUs
            n_species = len(species_set)
            # now add the unique OTU list to the output dictionary
            # if the key is not in the dict, create a new entry (= OTU ID plus number of OTUs)
            if i not in draw_dictionary.keys():
                draw_dictionary[i] = [n_species]
            # if the key already exists, calculate the sum of the already existing number of OTUs and the new number of OTUs
            else:
                # create a new list to store the current number of OTUs
                add_species_list = draw_dictionary[i]
                add_species_list.append(n_species)
                draw_dictionary[i] = add_species_list

            # remove the sample to draw only once
            sample_dict_to_draw.pop(random_choice)

        ############################################################################
        event, values = window_progress_bar.read(timeout=1)
        if event == 'Cancel'  or event is None:
            print('\nCancel')
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    # create a dict to store the average number of OTUs per draw
    rarefaction_dict_average, rarefaction_dict_stdef = {}, {}

    def average(lst):
        return sum(lst) / len(lst)

    # iterate through the draw_dictionary and calculate the average number of OTUs
    for key, value in draw_dictionary.items():
        average_species = average(draw_dictionary[key])
        stdef_species = np.std(draw_dictionary[key], dtype=np.float64)
        rarefaction_dict_average[key] = average_species
        rarefaction_dict_stdef[key] = stdef_species

    if error_style == "a":
        draws = [i+1 for i in rarefaction_dict_average.keys()] #list(rarefaction_dict_average.keys())
        n_species = list(rarefaction_dict_average.values())
        error_bar = list(rarefaction_dict_stdef.values())
        plt.figure(figsize=(20, 10))
        if rarefaction_ylim != '':
            plt.ylim(0, int(rarefaction_ylim))
        plt.errorbar(draws, n_species, error_bar, linewidth=0.8, color='blue', capsize=1.3, capthick=1, ecolor='lightgrey')
        plt.xticks(np.arange(1, len(draws)+1, step=1))
        plt.xlabel('# samples')
        plt.ylabel('# species')
        plt.title('repetitions = ' + str(replicates))


    elif error_style == "b":
        draws = [i+1 for i in rarefaction_dict_average.keys()] #list(rarefaction_dict_average.keys())
        n_species = list(rarefaction_dict_average.values())
        y = np.asarray(n_species)
        error_bar = np.asarray(list(rarefaction_dict_stdef.values()))
        plt.figure(figsize=(20, 10))
        if rarefaction_ylim != '':
            plt.ylim(0, int(rarefaction_ylim))
        plt.plot(draws, n_species)
        plt.fill_between(draws, y-error_bar, y+error_bar, alpha=0.1)
        plt.xticks(np.arange(1, len(draws)+1, step=1))
        plt.xlabel('# samples')
        plt.ylabel('# species')
        plt.title('repetitions = ' + str(replicates))


    plt.show(block=False)
    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
    if answer == "Yes":
        plt.savefig(output_file, bbox_inches='tight')
        plt.close()
        closing_text = "\n" + "Rarefaction curves are found in: " + str(path_to_outdirs) + "/rarefaction_curves/"
        print(closing_text)
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        from taxontabletools.create_log import ttt_log
        ttt_log("rarefaction curve species", "analysis", TaXon_table_file.name, output_file.name, "nan", path_to_outdirs)

    else:
        plt.close()



def CCA_analysis(TaXon_table_xlsx, meta_data_to_test, cca_w, cca_h, cca_s, path_to_outdirs):

    import pandas as pd
    from skbio.stats.ordination import cca
    import numpy as np
    import matplotlib.pyplot as plt
    from pathlib import Path
    import PySimpleGUI as sg

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    IDs_list = TaXon_table_df["IDs"].values.tolist()
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    output_pdf = Path(str(path_to_outdirs) + "/" + "CCA_plots" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + ".pdf")
    output_xlsx = Path(str(path_to_outdirs) + "/" + "CCA_plots" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + ".xlsx")

    # check for presence absence data
    # otherwise abort and print error message
    pa_test = set([val for sublist in TaXon_table_df[TaXon_table_samples].values.tolist() for val in sublist])
    if pa_test != {1,0}:
        print("Please use presence absence data!")
        sg.Popup("Please use presence absence data!", title=("Error"))
        raise RuntimeError

    # check for presence absence data
    # otherwise abort and print error message
    for i in Meta_data_table_df[meta_data_to_test]:
        if type(i) != int:
            print("Please use categorial numbers (ints) as meta data!")
            sg.Popup("Please use categorial numbers (ints) as meta data!", title=("Error"))
            raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == len(Meta_data_table_df['Samples'].tolist()):
        print("The meta data is unique for all samples. Please adjust the meta data table!")
        sg.Popup("The meta data is unique for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        print("The meta data is similar for all samples. Please adjust the meta data table!")
        sg.Popup("The meta data is similar for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    # check for empty samples and later remove them from the dataframe
    # otherwise abort and print error message
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    empty_samples_list = []
    for sample in TaXon_table_samples:
        empty_sample_test = set(TaXon_table_df[sample].values.tolist())
        if empty_sample_test == {0}:
            empty_samples_list.append(sample)
    if empty_samples_list != []:
        print("Please remove empty samples first!")
        sg.Popup("Please remove empty samples first!", title=("Error"))
        raise RuntimeError

    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):
        # transpose the pa table
        df_features = TaXon_table_df[TaXon_table_samples].transpose()
        df_features.index = Meta_data_table_samples
        df_features = df_features.rename_axis("a")

        # create a constrains table from the metadata table
        constrains_list = []
        for col in Meta_data_table_df[["Samples", meta_data_to_test]].values.tolist():
            constrains_list.append(float(col[1]))
        df_constrains = pd.DataFrame(constrains_list, Meta_data_table_samples, [meta_data_to_test])
        df_constrains.index = Meta_data_table_samples
        df_constrains = df_constrains.rename_axis("a")

    ordination_result = cca(df_features, df_constrains)
    ordination_result.sample_constraints

    #######################################################################################
    # create window to ask for CCA axis to test
    def slices(list, slice):
        for i in range(0, len(list), slice):
            yield list[i : i + slice]

    # collect the cca proportion explained values
    proportion_explained_list = []
    for i, cca_axis in enumerate(ordination_result.proportion_explained):
        proportion_explained_list.append("CCA" + str(i+1) + " (" + str(round(cca_axis* 100, 2)) + " %)")

    cca_axis_checkboxes = list(slices([sg.Checkbox(name, key=name, size=(15,1)) for name in proportion_explained_list], 4))

    cca_window_layout = [
                [sg.Text('Check two axis to be displayed')],
                [sg.Frame(layout = cca_axis_checkboxes, title = '')],
                [sg.Button('Plot', key='Plot'), sg.Checkbox("Invert axes", key="inverted_axis")],
                [sg.Button('Back')],
                ]

    cca_window = sg.Window('CCA axis', cca_window_layout, keep_on_top=True)

    while True:
        event, values = cca_window.read()

        if event == 'Plot':
            # collect the cca axis values
            axis_to_plot = [key for key,value in values.items() if value == True and "CCA" in key]
            # pass on only if two cca axes were checked
            if len(axis_to_plot) != 2:
                sg.Popup("Please choose exactly two CCA axes", title="Error", keep_on_top=True)
            else:
                if values["inverted_axis"] == True:
                    cat2 = axis_to_plot[0].split()[0]
                    cat1 =  axis_to_plot[1].split()[0]
                else:
                    cat1 = axis_to_plot[0].split()[0]
                    cat2 =  axis_to_plot[1].split()[0]

                df_cca = ordination_result.samples[[cat1, cat2]]
                df_cca.insert(2, "label", df_constrains[meta_data_to_test].values.tolist(), True)
                plt.figure(figsize=(int(cca_w), int(cca_h)))
                plt.grid(color='gray', alpha=0.1)
                groups = df_cca.groupby("label")
                for name, group in groups:
                    plt.plot(group[cat1], group[cat2], marker="o", linestyle="", label=name, ms=cca_s)
                plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
                plt.xlabel(cat1 + " (" + str(round(ordination_result.proportion_explained[cat1]* 100, 2)) + " %)")
                plt.ylabel(cat2 + " (" + str(round(ordination_result.proportion_explained[cat2]* 100, 2)) + " %)")
                plt.title(meta_data_to_test)

                plt.show(block=False)
                answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
                if answer == "Yes":
                    plt.savefig(output_pdf)
                    plt.close()
                    ordination_result.samples[[cat1, cat2]].to_excel(output_xlsx)
                    closing_text = "Taxonomic resolution plots are found in: " + str(path_to_outdirs) + "/Taxonomic_resolution_plots/"
                    print(closing_text)
                    sg.Popup(closing_text, title="Finished", keep_on_top=True)

                    from taxontabletools.create_log import ttt_log
                    ttt_log("cca analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)

                    break
                else:
                    plt.close()

        if event == 'Back':
            break

    cca_window.close()















#



def calculate_taxonomic_richness(TaXon_table_xlsx, path_to_outdirs, x_tax_rich, y_tax_rich, font_tax_rich, ylim_tax_rich):

    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    import matplotlib.pyplot as plt
    from pathlib import Path

    TaXon_table_file =  Path(TaXon_table_xlsx)

    print("\n" + "Input file:", TaXon_table_file.name)

    taxonomic_levels = ["Phylum", "Class", "Order", "Family", "Genus", "Species"]

    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)
    statistics_list, statistics_set, statistics_dict, highest_level_dict = [], [], {}, {}
    writer = pd.ExcelWriter(TaXon_table_xlsx, engine = 'xlsxwriter')

    for taxon_to_evaluate in taxonomic_levels:
        taxa_list = [x for x in df[taxon_to_evaluate].values.tolist() if str(x) != 'nan']
        statistics = taxon_to_evaluate, len(taxa_list)
        statistics_set.append(len(set(taxa_list)))
        statistics_list.append(list(statistics))
        statistics_dict[taxon_to_evaluate] = len(taxa_list)

    highest_level_dict["Phylum"] = statistics_dict["Phylum"] - statistics_dict["Class"]
    highest_level_dict["Class"] = statistics_dict["Class"] - statistics_dict["Order"]
    highest_level_dict["Order"] = statistics_dict["Order"] - statistics_dict["Family"]
    highest_level_dict["Family"] = statistics_dict["Family"] - statistics_dict["Genus"]
    highest_level_dict["Genus"] = statistics_dict["Genus"] - statistics_dict["Species"]
    highest_level_dict["Species"] = statistics_dict["Species"]

    taxon_levels = list(highest_level_dict.keys())
    number_of_taxa_per_level = statistics_set

    # create plot
    output_file = Path(str(path_to_outdirs) + "/" + "Taxonomic_richness_plots" + "/" + TaXon_table_file.stem + "_taxonomic_richness_bar.pdf")

    plt.figure(figsize=(int(x_tax_rich), int(y_tax_rich)))
    if ylim_tax_rich != '':
        plt.ylim(0, int(ylim_tax_rich))
    plt.bar(taxon_levels, number_of_taxa_per_level)
    plt.ylabel('# taxa')
    plt.title('Number of taxa per taxonomic level')
    for i, v in enumerate(number_of_taxa_per_level):
        plt.text(i - 0.01, v, str(v), horizontalalignment='center', verticalalignment='center', fontsize=font_tax_rich)

    plt.show(block=False)
    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
    if answer == "Yes":
        plt.savefig(output_file, bbox_inches='tight')
        plt.close()

        closing_text = "\n" + "Taxonomic richness plots are found in: " + str(path_to_outdirs) + "/taxonomic_richness_plots/"
        print(closing_text)
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        from taxontabletools.create_log import ttt_log
        ttt_log("taxonomic richness", "analysis", TaXon_table_file.name, output_file.name, "nan", path_to_outdirs)

    else:
        plt.close()
